#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <gtk/gtk.h>
#include <glade/glade.h>

#include "elist.h"
#include "str_util.h"
#include "main_win.h"
#include "dir_util.h"
#include "audio_file.h"
#include "edit_tab.h"


/* widgets we need */
static GtkCList *clist_files = NULL;
static GtkNotebook *nb_file = NULL;
static GtkNotebook *nb_edit = NULL;
static GtkLabel *lab_info_names = NULL;
static GtkLabel *lab_info_values = NULL;
static GtkWidget *align_create_tag = NULL;
static GtkWidget *align_remove_tag = NULL;

/* private data */
static gchar *file_name;
static audio_file *af = NULL;


/*** private functions ******************************************************/

static void load_file(const gchar *name)
{
	GString *info_names;
	GString *info_values;
	gchar *buf, *aux, *tok;
	int res;

	/* in some cases the previous file may still be loaded */
	if (af) {
		audio_file_delete(af);
		af = NULL;
	}

	/* create the audio_file object */
	res = audio_file_new(&af, name, TRUE);
	if (res == AF_ERR_FILE)
		res = audio_file_new(&af, name, FALSE);

	if (res != AF_OK) {
		if (res == AF_ERR_FILE)
			mw_status_printf("Couldn't open file %s", name);
		else if (res == AF_ERR_FORMAT)
			mw_status_printf("Audio format not recognized: %s", name);
		else 
			mw_status_printf("Unknow error when opening file %s", name);

		return;
	}

	if (audio_file_is_editable(af))
		mw_status_clear();
	else
		mw_status_printf("File is read-only: %s", name);

	
	/* fill in the File Info box */
	info_names = g_string_sized_new(128);
	info_values = g_string_sized_new(128);
	
	g_string_append(info_names, "File Type:");
	g_string_append(info_values, audio_file_get_desc(af));

	buf = strdup(audio_file_get_info(af));
	aux = buf;
	while(1) {
		tok = strtok(aux, "\n");
		if (!tok) break;
		g_string_sprintfa(info_names, "\n%s:", tok);
		if (aux) aux = NULL;

		tok = strtok(aux, "\n");
		if (!tok) break;
		g_string_sprintfa(info_values, "\n%s", tok);
	}

	gtk_label_set_text(lab_info_names, info_names->str);
	gtk_label_set_text(lab_info_values, info_values->str);

	free(buf);
	g_string_free(info_names, TRUE);
	g_string_free(info_values, TRUE);


	/* display the appropriate edit interface */
	audio_file_edit_load(af);

	gtk_notebook_set_page(nb_file, 1);
}

/* UI callbacks */
void cb_notebook_switch(GtkNotebook *nb, GtkNotebookPage *page, gint page_num, gpointer user_data)
{
	static gint last = 0;

	/* when leaving edit tab */
	if (last == 0 && page_num != 0)
		gtk_clist_unselect_all(clist_files);

	last = page_num;
}

void cb_selection_changed(GtkCList *clist, gint row, gint col, GdkEventButton *event, gpointer data)
{
	/* Because of the mess that GtkCList makes of emmiting the select_row and 
	   unselect_row signals, the only sane way of doing this is to track the 
	   selection state ourselves.
	   Unfortunately, because of the mess GtkCList makes of its API, there is 
	   no way to find out which rows are selected, so we have to resort to using 
	   the private fields of GtkCList.  Sigh. */

	gchar *new_name;

	if (clist->selection == NULL ||			/* no selection */
	    clist->selection != clist->selection_end)	/* multiple selections */
	{
		if (file_name != NULL) {
			et_unload_file();
			file_name = NULL;
		}
		return;
	}

	new_name = gtk_clist_get_row_data(clist, (int)clist->selection->data);

	if (new_name == NULL)
		return;
	if (file_name != NULL && strcmp(new_name, file_name) == 0)
		return;

	file_name = new_name;
	load_file(file_name);
}


/*** public functions *******************************************************/

void et_init(GladeXML *xml)
{
	clist_files = GTK_CLIST(glade_xml_get_widget(xml, "clist_files"));
	nb_file = GTK_NOTEBOOK(glade_xml_get_widget(xml, "nb_file"));
	nb_edit = GTK_NOTEBOOK(glade_xml_get_widget(xml, "nb_edit"));
	align_create_tag = glade_xml_get_widget(xml, "align_create_tag");
	align_remove_tag = glade_xml_get_widget(xml, "align_remove_tag");
	lab_info_names = GTK_LABEL(glade_xml_get_widget(xml, "lab_info_names"));
	lab_info_values = GTK_LABEL(glade_xml_get_widget(xml, "lab_info_values"));
}


void et_advance_file()
{
	/* The only way to get the current list selection is through the
	   private GtkCList fields selection and selection_end */

	int new_sel;

	if (clist_files->selection == NULL)
		return;

	new_sel = (int)clist_files->selection_end->data + 1;

	gtk_clist_unselect_all(clist_files);
	gtk_clist_select_row(clist_files, new_sel, 0);
}


void et_unload_file()
{
	if (af) {
		audio_file *temp = af;
		af = NULL;
		audio_file_edit_unload(temp);
		audio_file_delete(temp);

		gtk_notebook_set_page(nb_file, 0);
	}
}


void et_update_file_list(const GEList *file_list)
{
	gchar *last_file = "";
	gchar *aux;
	gint row;
	GtkStyle* style;
	GList* iter;

	gtk_clist_clear(clist_files);

	if (file_list == NULL)
		return;

	style = gtk_widget_get_style(GTK_WIDGET(clist_files));
	if (style == NULL)
		g_warning("Couldn't get style for widget clist_files");

	gtk_clist_freeze(clist_files);

	iter = g_elist_first(file_list);
	while (iter) {
		if (!du_compare_file_paths(last_file, iter->data)) {
			aux = g_dirname(iter->data);
			row = gtk_clist_append(clist_files, (char **)&aux);
			gtk_clist_set_row_data(clist_files, row, NULL);
			gtk_clist_set_selectable(clist_files, row, FALSE);
			if (style) {
				gtk_clist_set_background(clist_files, row, &style->bg[0]);
				gtk_clist_set_foreground(clist_files, row, &style->fg[0]);
			}
			g_free(aux);
		}
		aux = g_basename(iter->data);
		row = gtk_clist_append(clist_files, (char **)&aux);
		gtk_clist_set_row_data(clist_files, row, iter->data);
		last_file = iter->data;
		iter = g_list_next(iter);
	}
	gtk_clist_columns_autosize(clist_files);

	gtk_clist_thaw(clist_files);
}

