#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <gtk/gtk.h>
#include <glade/glade.h>

#include "elist.h"
#include "str_util.h"
#include "dir_util.h"
#include "mru.h"
#include "prefs.h"
#include "audio_file.h"
#include "dir_select.h"
#include "char_conv_dlg.h"
#include "progress_msg.h"
#include "cursor.h"
#include "help.h"

#include "rename_tab.h"


/* widgets we need */
static GtkCombo *combo_rename_format = NULL;
static GtkEntry *ent_rename_format = NULL;
static GtkButton *b_rename_edit_format = NULL;
static GtkButton *b_rename_go = NULL;

/* preferences */
static MRUList *format_mru;


/*** private functions ******************************************************/

static void append_field(GString *gstr, audio_file *af, int field)
{
	const char *value;
	char *aux;

	audio_file_get_field(af, field, &value);
	aux = strdup(value);

	/* apply character conversions */
	chconv_rename_options options = chconv_get_rename_options();
	if (options.space_conv != NULL) {
		str_replace_char(aux, ' ', options.space_conv[0]);
	}
	if (options.invalid_conv != NULL) {
		str_replace_char(aux, '/', options.invalid_conv[0]);
	} else {
		str_remove_char(aux, '/');
	}
	if (options.case_conv != CASE_CONV_NONE) {
		str_convert_case(aux, options.case_conv);
	}

	/* append to name */
	if (field == AF_TRACK) {
		long track;
		char *endptr;
		track = strtol(aux, &endptr, 10);
		if (*endptr == 0)
			g_string_sprintfa(gstr, "%02li", track);
		else
			g_string_append(gstr, aux);
	} else {
		g_string_append(gstr, aux);
	}
	free(aux);
}

static void build_file_name(gchar *format, audio_file *af, GString *new_name)
{
	gchar *p;
	gint i, span;

	/* fill in the filename from the tag according to the given format */
	i = 0;
	while (TRUE) {
		p = index(&format[i], '<');
		if (p != NULL) {
			span = (gint)(p - &format[i]);
			if (span > 0) {
				g_string_sprintfa(new_name, "%.*s", span, &format[i]);
				i += span;
			} else if (strncmp(&format[i], "<title>", 7) == 0) {
				append_field(new_name, af, AF_TITLE);
				i += 7;
			} else if (strncmp(&format[i], "<artist>", 8) == 0) {
				append_field(new_name, af, AF_ARTIST);
				i += 8;
			} else if (strncmp(&format[i], "<album>", 7) == 0) {
				append_field(new_name, af, AF_ALBUM);
				i += 7;
			} else if (strncmp(&format[i], "<year>", 6) == 0) {
				append_field(new_name, af, AF_YEAR);
				i += 6;
			} else if (strncmp(&format[i], "<comment>", 9) == 0) {
				append_field(new_name, af, AF_COMMENT);
				i += 9;
			} else if (strncmp(&format[i], "<track>", 7) == 0) {
				append_field(new_name, af, AF_TRACK);
				i += 7;
			} else if (strncmp(&format[i], "<genre>", 7) == 0) {
				append_field(new_name, af, AF_GENRE);
				i += 7;
			} else {
				g_string_append_c(new_name, format[i++]);
			}
		} else {
			g_string_append(new_name, &format[i]);
			break;
		}
	}

	/* add back the extension */
	g_string_append(new_name, audio_file_get_extension(af));
}


static void rename_files() 
{
	GEList *file_list;
	GList *iter;
	audio_file *af = NULL;
	gchar *format;
	gchar *last_full_name = "";
	gchar *orig_full_name, *orig_name;
	GString *new_full_name, *new_path;
	gchar *p;
	gint new_dirs;
	gint count_total, count_renamed;
	gboolean moving;
	struct stat st_info;
	int res, save_errno;

	new_full_name = g_string_sized_new(256);
	new_path = g_string_sized_new(256);

	format = gtk_entry_get_text(ent_rename_format);
	moving = (strchr(format, '/') ? TRUE : FALSE);

	pm_start(PM_RENAME_PROGRESS);
	pm_printf(PM_RENAME_PROGRESS, PM_ICON_INFO, "Starting in directory \"%s\"", ds_get_working_dir());

	file_list = ds_get_file_list();
	count_total = 0;
	count_renamed = 0;
	for (iter = g_elist_first(file_list); iter; iter = g_list_next(iter)) {
		/* flush pending gtk operations so the UI doesn't freeze */
		pm_scroll_to_bottom(PM_RENAME_PROGRESS);
		while (gtk_events_pending()) gtk_main_iteration();
		
		count_total++;

		orig_full_name = (gchar *)iter->data;
		p = strrchr(orig_full_name, '/');
		if (!du_compare_file_paths(last_full_name, orig_full_name))
			pm_printf(PM_RENAME_PROGRESS, PM_ICON_INFO, "Entering directory \"%.*s\"", (gint)(p-orig_full_name), orig_full_name);
		last_full_name = orig_full_name;

		/* read the file information */
		orig_name = g_basename(orig_full_name);
		res = audio_file_new(&af, orig_full_name, FALSE);
		if (res != AF_OK) {
			pm_printf(PM_RENAME_PROGRESS, PM_ICON_FAIL, "Error renaming \"%s\"", orig_name);

			if (res == AF_ERR_FILE)
				pm_printf(PM_RENAME_PROGRESS, PM_ICON_NONE, "Couldn't open file for reading");
			else if (res == AF_ERR_FORMAT)
				pm_printf(PM_RENAME_PROGRESS, PM_ICON_NONE, "Audio format not recognized");
			else 
				pm_printf(PM_RENAME_PROGRESS, PM_ICON_NONE, "Unknown error");

			goto cont;
		}

		if (!audio_file_has_tag(af)) {
			pm_printf(PM_RENAME_PROGRESS, PM_ICON_FAIL, "Error renaming \"%s\"", orig_name);
			pm_printf(PM_RENAME_PROGRESS, PM_ICON_NONE, "File has no tag");
			goto cont;
		}

		/* build the new file name (with path) */
		g_string_truncate(new_full_name, 0);
		if (p) 
			g_string_sprintfa(new_full_name, "%.*s/", (gint)(p-orig_full_name), orig_full_name);
		build_file_name(format, af, new_full_name);

		if (strcmp(orig_full_name, new_full_name->str) == 0) {
			pm_printf(PM_RENAME_PROGRESS, PM_ICON_OK, "File name \"%s\" already in desired format", orig_name);
			goto cont;
		}

		if (!stat(new_full_name->str, &st_info)) {
			pm_printf(PM_RENAME_PROGRESS, PM_ICON_FAIL, "Error %s \"%s\"", (moving ? "moving" : "renaming"), orig_name);
			pm_printf(PM_RENAME_PROGRESS, PM_ICON_NONE, "File \"%s\" already exists", new_full_name->str);
			goto cont;
		}

		/* create the destination dir if necessary */
		p = strrchr(new_full_name->str, '/');
		if (moving && p) {
			new_dirs = 0;
			g_string_sprintf(new_path, "%.*s/", (gint)(p-new_full_name->str), new_full_name->str);
			res = du_make_dir_tree(new_path->str, &new_dirs);
			if (!res) {
				save_errno = errno;
				pm_printf(PM_RENAME_PROGRESS, PM_ICON_FAIL, "Error creating directory \"%s\"", new_path->str);
				pm_printf(PM_RENAME_PROGRESS, PM_ICON_NONE, "%s (errno %d)", strerror(save_errno), save_errno);
				goto cont;
			}
			if (new_dirs) {
				pm_printf(PM_RENAME_PROGRESS, PM_ICON_OK, "Created directory \"%s\"", new_path->str);
			}
		}

		/* rename the file */
		res = rename(orig_full_name, new_full_name->str);
		if (res != 0) {
			save_errno = errno;
			pm_printf(PM_RENAME_PROGRESS, PM_ICON_FAIL, "Error %s \"%s\"", (moving ? "moving" : "renaming"), orig_name);
			pm_printf(PM_RENAME_PROGRESS, PM_ICON_NONE, "%s (errno %d)", strerror(save_errno), save_errno);
		} else {
			if (moving)
				pm_printf(PM_RENAME_PROGRESS, PM_ICON_OK, "Moved \"%s\" to \"%s\"", orig_name, new_full_name->str);
			else
				pm_printf(PM_RENAME_PROGRESS, PM_ICON_OK, "Renamed \"%s\" to \"%s\"", orig_name, g_basename(new_full_name->str));
			count_renamed++;
		}

	cont:
		if (af) {
			audio_file_delete(af);
			af = NULL;
		}
	}

	g_string_free(new_full_name, TRUE);
	g_string_free(new_path, TRUE);

	pm_printf(PM_RENAME_PROGRESS, PM_ICON_INFO, "Done (%s %d of %d files)", 
		  (moving ? "moved" : "renamed"), count_renamed, count_total);
	pm_scroll_to_bottom(PM_RENAME_PROGRESS);
	pm_end(PM_RENAME_PROGRESS);

	if (count_renamed > 0)
		ds_set_working_dir(NULL);
}


/* UI callbacks */
void cb_rename_go(GtkButton *button, gpointer user_data)
{
	GEList *list = ds_get_file_list();
	if (g_elist_length(list) == 0) {
		pm_start(PM_RENAME_PROGRESS);
		pm_printf(PM_RENAME_PROGRESS, PM_ICON_FAIL, "No files selected");
		pm_end(PM_RENAME_PROGRESS);
		return;
	}

	mru_add(format_mru, gtk_entry_get_text(ent_rename_format));
	gtk_combo_set_popdown_strings(combo_rename_format, GLIST(format_mru->list));

	cursor_set_wait();
	rename_files();
	cursor_set_normal();
}

void cb_show_rename_chconv(GtkButton *button, gpointer user_data)
{
	chconv_display(CHCONV_RENAME);
}

void cb_rename_help(GtkButton *button, gpointer user_data)
{
	help_display(HELP_RENAME_FORMAT);
}


/*** public functions *******************************************************/

void rt_init(GladeXML *xml)
{
	GEList *format_list;

	/*
	 * get the widgets from glade
	 */

	combo_rename_format = GTK_COMBO(glade_xml_get_widget(xml, "combo_rename_format"));
	ent_rename_format = GTK_ENTRY(glade_xml_get_widget(xml, "ent_rename_format"));
	b_rename_edit_format = GTK_BUTTON(glade_xml_get_widget(xml, "b_rename_edit_format"));
	b_rename_go = GTK_BUTTON(glade_xml_get_widget(xml, "b_rename_go"));

	/*
	 * get the preference values, or set them to defaults
	 */

	/* format_mru */
	format_list = pref_get_ref("rt:format_mru");
	if (!format_list) {
		GEList *temp_list = g_elist_new();
		g_elist_append(temp_list, "<track>. <title>");
		g_elist_append(temp_list, "<artist> - <title>");
		g_elist_append(temp_list, "<artist> - <album>/<track>. <title>");
		g_elist_append(temp_list, "<artist>/<album>/<track>. <title>");
		g_elist_append(temp_list, "<artist>/<album> (<year>)/<track>. <title>");
		format_list = pref_set("rt:format_mru", PREF_STRING | PREF_LIST, temp_list);
		g_elist_free(temp_list);
	}
	format_mru = mru_new_from_list(10, format_list);

	gtk_combo_set_popdown_strings(combo_rename_format, GLIST(format_mru->list));
}

