#include <config.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <gtk/gtk.h>
#include <glade/glade.h>

#ifdef HAVE_GNUREGEX_H
# include <gnuregex.h>	/* used in FreeBSD, I'm told */
#elif HAVE_REGEX_H
# include <regex.h>	/* for everyone else */
#endif

#include "elist.h"
#include "str_util.h"
#include "dir_util.h"
#include "mru.h"
#include "prefs.h"
#include "audio_file.h"
#include "genre.h"
#include "dir_select.h"
#include "char_conv_dlg.h"
#include "modal_dialog.h"
#include "progress_msg.h"
#include "cursor.h"
#include "help.h"
#include "tag_tab.h"


/* for parsing file names */
typedef struct {
	/* regexp to match against the file name */
	regex_t re;
	/* indexes of substring matches */
	gint title;
	gint artist;
	gint album;
	gint year;
	gint comment;
	gint track;
	gint genre;
} parse_info;


/* widgets we need */
static GtkCombo *combo_tag_format = NULL;
static GtkEntry *ent_tag_format = NULL;
static GtkButton *b_tag_edit_format = NULL;
static GtkCheckButton *cb_use_filename = NULL;
static GtkCheckButton *cb_remove_tags = NULL;
static GtkButton *b_tag_go = NULL;
static GtkCheckButton *cb_title = NULL;
static GtkCheckButton *cb_artist = NULL;
static GtkCheckButton *cb_album = NULL;
static GtkCheckButton *cb_year = NULL;
static GtkCheckButton *cb_comment = NULL;
static GtkCheckButton *cb_track = NULL;
static GtkCheckButton *cb_genre = NULL;
static GtkEntry *ent_title = NULL;
static GtkEntry *ent_artist = NULL;
static GtkEntry *ent_album = NULL;
static GtkEntry *ent_year = NULL;
static GtkEntry *ent_comment = NULL;
static GtkEntry *ent_track = NULL;
static GtkCheckButton *cb_track_auto = NULL;
static GtkCombo *combo_genre = NULL;
static GtkEntry *ent_genre = NULL;
static GtkWindow *w_main = NULL;
static GtkWidget *clist_tag_progress = NULL;

/* private data */
static gboolean ignore_field_changed = FALSE;
static gboolean from_fn_title = FALSE;
static gboolean from_fn_artist = FALSE;
static gboolean from_fn_album = FALSE;
static gboolean from_fn_year = FALSE;
static gboolean from_fn_comment = FALSE;
static gboolean from_fn_track = FALSE;
static gboolean from_fn_genre = FALSE;

/* preferences */
static gboolean *use_filename;
static MRUList *format_mru;


/*** private functions ******************************************************/

static parse_info *build_parse_info(gchar *format)
{
	parse_info *info = calloc(1, sizeof(parse_info));
	gchar *aux_str = calloc(1, 2 + 2*strlen(format));
	gchar *expr_str = calloc(1, 2 + 2*strlen(format));
	gint i, j, span, pos;
	gchar *p;
	gint res;


	/* printf("format: %s\n", format); */

	/* constrain the expression to match the whole string and escape
	   characters that have special meaning in regexps */
	i = 0;
	j = 0;
	aux_str[j++] = '^';
	while (format[i]) {
		if (strchr(".*+?[]{}()|^$\\", format[i]))
			aux_str[j++] = '\\';
		aux_str[j++] = format[i++];
	}
	aux_str[j++] = '$';
	aux_str[j++] = 0;

	/* find the tag fields, record their relative positions and replace 
	   them in the final expression */
	i = 0;
	j = 0;
	pos = 1;
	while (TRUE) {
		p = index(&aux_str[i], '<');
		if (p != NULL) {
			span = (gint)(p - &aux_str[i]);
			if (span > 0) {
				strncpy(&expr_str[j], &aux_str[i], span);
				i += span;
				j += span;
			} else if (strncmp(&aux_str[i], "<title>", 7) == 0) {
				strncpy(&expr_str[j], "(.*)", 4);
				info->title = pos++;
				i += 7;
				j += 4;
			} else if (strncmp(&aux_str[i], "<artist>", 8) == 0) {
				strncpy(&expr_str[j], "(.*)", 4);
				info->artist = pos++;
				i += 8;
				j += 4;
			} else if (strncmp(&aux_str[i], "<album>", 7) == 0) {
				strncpy(&expr_str[j], "(.*)", 4);
				info->album = pos++;
				i += 7;
				j += 4;
			} else if (strncmp(&aux_str[i], "<year>", 6) == 0) {
				strncpy(&expr_str[j], "(.*)", 4);
				info->year = pos++;
				i += 6;
				j += 4;
			} else if (strncmp(&aux_str[i], "<comment>", 9) == 0) {
				strncpy(&expr_str[j], "(.*)", 4);
				info->comment = pos++;
				i += 9;
				j += 4;
			} else if (strncmp(&aux_str[i], "<track>", 7) == 0) {
				strncpy(&expr_str[j], "([0-9]*)", 8);
				info->track = pos++;
				i += 7;
				j += 8;
			} else if (strncmp(&aux_str[i], "<genre>", 7) == 0) {
				strncpy(&expr_str[j], "(.*)", 4);
				info->genre = pos++;
				i += 7;
				j += 4;
			} else if (strncmp(&aux_str[i], "<\\*>", 4) == 0) {
				strncpy(&expr_str[j], ".*", 2);
				i += 4;
				j += 2;
			} else {
				expr_str[j++] = aux_str[i++];
			}
		} else {
			strcpy(&expr_str[j], &aux_str[i]);
			break;
		}
	}


	/* printf("regexp: %s\n", expr_str); */

	/* compile the regexp from the expression string */
	res = regcomp(&info->re, expr_str, REG_EXTENDED | REG_ICASE);
	if (res != 0) {
		g_warning("error in regcomp (\"%s\"): %d", expr_str, res);
		free(aux_str);
		free(expr_str);
		free(info);
		return NULL;
	}

	free(aux_str);
	free(expr_str);

	return info;
}

static void free_parse_info(parse_info *info)
{
	regfree(&info->re);
	free(info);
}


static void set_field(audio_file *af, int field, gchar *src, int len)
{
	static int buf_size = 256;
	static gchar *buf = NULL;
	if (buf == NULL)
		buf = malloc(buf_size);
	if (len >= buf_size) {
		buf_size = len + 1;
		buf = realloc(buf, buf_size);
	}

	strncpy(buf, src, len);
	buf[len] = 0;

	/* apply character conversions */
	chconv_tag_options options = chconv_get_tag_options();
	if (options.space_conv != NULL) {
		str_replace_chars(buf, options.space_conv, ' ');
		str_trim(buf, ' ');
	}
	if (options.case_conv != CASE_CONV_NONE) {
		str_convert_case(buf, options.case_conv);
	}

	audio_file_set_field(af, field, buf);
}

static gboolean tag_from_filename(parse_info *pi, gchar *filename, audio_file *af)
{
	regmatch_t matches[8];
	char *p;
	int res;

	/* temporarily truncate the file extension */
	p = strrchr(filename, '.');
	if (p && strlen(p) <= 5)
		*p = 0;
	else
		p = NULL;

	/* printf("string: %s\n", filename); */

	
	/* match the regexp against the file name (minus extension) and 
	   extract the tag fields */
	res = regexec(&pi->re, filename, 8, matches, 0);
	if (res != 0) {
		if (p) *p = '.';
		return FALSE;
	}

	if (pi->title > 0) {
		set_field(af, AF_TITLE, &filename[matches[pi->title].rm_so], 
			  matches[pi->title].rm_eo - matches[pi->title].rm_so);
	}
	if (pi->artist > 0) {
		set_field(af, AF_ARTIST, &filename[matches[pi->artist].rm_so], 
			  matches[pi->artist].rm_eo - matches[pi->artist].rm_so);
	}
	if (pi->album > 0) {
		set_field(af, AF_ALBUM, &filename[matches[pi->album].rm_so], 
			  matches[pi->album].rm_eo - matches[pi->album].rm_so);
	}
	if (pi->year > 0) {
		set_field(af, AF_YEAR, &filename[matches[pi->year].rm_so], 
			  matches[pi->year].rm_eo - matches[pi->year].rm_so);
	}
	if (pi->comment > 0) {
		set_field(af, AF_COMMENT, &filename[matches[pi->comment].rm_so], 
			  matches[pi->comment].rm_eo - matches[pi->comment].rm_so);
	}
	if (pi->track > 0) {
		set_field(af, AF_TRACK, &filename[matches[pi->track].rm_so], 
			  matches[pi->track].rm_eo - matches[pi->track].rm_so);
	}
	if (pi->genre > 0) {
		set_field(af, AF_GENRE, &filename[matches[pi->genre].rm_so], 
			  matches[pi->genre].rm_eo - matches[pi->genre].rm_so);
	}

	/* put the extension back */
	if (p) *p = '.';

	return TRUE;
}

static void clear_tags() 
{
	GEList *file_list;
	GList *iter;
	audio_file *af;
	gchar *last_path = "";
	gchar *curr_path, *p;
	int count_total, count_tagged;
	int save_errno, res;

	pm_start(PM_TAG_PROGRESS);
	pm_printf(PM_TAG_PROGRESS, PM_ICON_INFO, "Starting in directory \"%s\"", ds_get_working_dir());

	file_list = ds_get_file_list();
	count_total = 0;
	count_tagged = 0;
	for (iter = g_elist_first(file_list); iter; iter = g_list_next(iter)) {
		/* flush pending gtk operations so the UI doesn't freeze */
		pm_scroll_to_bottom(PM_TAG_PROGRESS);
		while (gtk_events_pending()) gtk_main_iteration();

		count_total++;

		curr_path = (gchar *)iter->data;
		p = strrchr(curr_path, '/');

		if (!du_compare_file_paths(last_path, curr_path))
			pm_printf(PM_TAG_PROGRESS, PM_ICON_INFO, "Entering directory \"%.*s\"", (gint)(p-curr_path), curr_path);
		last_path = curr_path;

		res = audio_file_new(&af, curr_path, TRUE);
		if (res != AF_OK) {
			pm_printf(PM_TAG_PROGRESS, PM_ICON_FAIL, "Error in file \"%s\"", g_basename(curr_path));

			if (res == AF_ERR_FILE)
				pm_printf(PM_TAG_PROGRESS, PM_ICON_NONE, "Couldn't open file for writing");
			else if (res == AF_ERR_FORMAT)
				pm_printf(PM_TAG_PROGRESS, PM_ICON_NONE, "Audio format not recognized");
			else 
				pm_printf(PM_TAG_PROGRESS, PM_ICON_NONE, "Unknown error (%d)", res);

			goto cont;
		}

		audio_file_remove_tag(af);
		res = audio_file_write_changes(af);
		if (res != AF_OK) {
			save_errno = errno;
			pm_printf(PM_TAG_PROGRESS, PM_ICON_FAIL, "Error in file \"%s\"", g_basename(curr_path));
			pm_printf(PM_TAG_PROGRESS, PM_ICON_NONE, "%s (errno %d)", strerror(save_errno), save_errno);
		}

		pm_printf(PM_TAG_PROGRESS, PM_ICON_OK, "Cleared tag from \"%s\"", g_basename(curr_path));
		count_tagged++;

	cont:
		if (af) {
			audio_file_delete(af);
			af = NULL;
		}
	}

	pm_printf(PM_TAG_PROGRESS, PM_ICON_INFO, "Done (cleared %d of %d files)", count_tagged, count_total);
	pm_scroll_to_bottom(PM_TAG_PROGRESS);
	pm_end(PM_TAG_PROGRESS);
}

static void tag_files() 
{
	GEList *file_list;
	GList *iter;
	audio_file *af;
	parse_info *pi;
	gint path_components;
	gchar *last_path;
	gchar *curr_path;
	gchar *file_name;
	gint track_auto_index;
	gint count_total, count_tagged;
	int save_errno, res;

	gboolean write_title   = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(cb_title));
	gboolean write_artist  = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(cb_artist));
	gboolean write_album   = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(cb_album));
	gboolean write_year    = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(cb_year));
	gboolean write_comment = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(cb_comment));
	gboolean write_track   = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(cb_track));
	gboolean write_genre   = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(cb_genre));
	gboolean track_auto    = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(cb_track_auto));

	pm_start(PM_TAG_PROGRESS);
	if (!(write_title || write_artist || write_album || write_year || 
	      write_comment || write_track || write_genre))
	{
		pm_printf(PM_TAG_PROGRESS, PM_ICON_INFO, "No tag fields to set!");
		pm_end(PM_TAG_PROGRESS);
		return;
	}
	pm_printf(PM_TAG_PROGRESS, PM_ICON_INFO, "Starting in directory \"%s\"", ds_get_working_dir());

	if (*use_filename) {
		gchar *format = gtk_entry_get_text(ent_tag_format);
		pi = build_parse_info(format);
		path_components = du_count_path_components(format);
	}
	else {
		pi = NULL;
		path_components = 0;
	}

	file_list = ds_get_file_list();
	count_total = 0;
	count_tagged = 0;
	track_auto_index = 1;
	last_path = "";
	for (iter = g_elist_first(file_list); iter; iter = g_list_next(iter)) {
		/* flush pending gtk operations so the UI doesn't freeze */
		pm_scroll_to_bottom(PM_TAG_PROGRESS);
		while (gtk_events_pending()) gtk_main_iteration();

		count_total++;

		curr_path = (gchar *)iter->data;

		if (!du_compare_file_paths(last_path, curr_path)) {
			gchar *p = strrchr(curr_path, '/');
			pm_printf(PM_TAG_PROGRESS, PM_ICON_INFO, "Entering directory \"%.*s\"", (gint)(p-curr_path), curr_path);
			track_auto_index = 1; /* reset track auto-increment */
		}
		last_path = curr_path;

		file_name = g_basename(curr_path);
		res = audio_file_new(&af, curr_path, TRUE);
		if (res != AF_OK) {
			pm_printf(PM_TAG_PROGRESS, PM_ICON_FAIL, "Error tagging \"%s\"", file_name);

			if (res == AF_ERR_FILE)
				pm_printf(PM_TAG_PROGRESS, PM_ICON_NONE, "Couldn't open file for writing");
			else if (res == AF_ERR_FORMAT)
				pm_printf(PM_TAG_PROGRESS, PM_ICON_NONE, "Audio format not recognized");
			else 
				pm_printf(PM_TAG_PROGRESS, PM_ICON_NONE, "Unknown error (%d)", res);

			goto cont;
		}

		/* create the tag if not already present */
		audio_file_create_tag(af);
	
		/* fill in the values from the form */
		if (write_title && !(from_fn_title && *use_filename))
			audio_file_set_field(af, AF_TITLE, gtk_entry_get_text(ent_title));
		if (write_artist && !(from_fn_artist && *use_filename))
			audio_file_set_field(af, AF_ARTIST, gtk_entry_get_text(ent_artist));
		if (write_album && !(from_fn_album && *use_filename))
			audio_file_set_field(af, AF_ALBUM, gtk_entry_get_text(ent_album));
		if (write_year && !(from_fn_year && *use_filename))
			audio_file_set_field(af, AF_YEAR, gtk_entry_get_text(ent_year));
		if (write_comment && !(from_fn_comment && *use_filename))
			audio_file_set_field(af, AF_COMMENT, gtk_entry_get_text(ent_comment));
		if (write_track && !(from_fn_track && *use_filename)) {
			if (track_auto) {
				char buf[10];
				snprintf(buf, 10, "%02u", track_auto_index++);
				audio_file_set_field(af, AF_TRACK, buf);
			} else 
				audio_file_set_field(af, AF_TRACK, gtk_entry_get_text(ent_track));
		}
		if (write_genre && !(from_fn_genre && *use_filename))
			audio_file_set_field(af, AF_GENRE, gtk_entry_get_text(GTK_ENTRY(combo_genre->entry)));

		/* fill in the values from the file name */
		if (pi != NULL) {
			gchar *full_path = du_join_path(ds_get_working_dir(), curr_path);
			gchar *relevant_path = du_last_n_path_components(full_path, path_components);

			res = tag_from_filename(pi, relevant_path, af);
			free(full_path);

			if (!res) {
				pm_printf(PM_TAG_PROGRESS, PM_ICON_FAIL, "Error tagging \"%s\"", file_name);
				pm_printf(PM_TAG_PROGRESS, PM_ICON_NONE, "File name does not match expected format");
				goto cont;
			}
		}

		/* write new tag to file */
		res = audio_file_write_changes(af);
		if (res != AF_OK) {
			save_errno = errno;
			pm_printf(PM_TAG_PROGRESS, PM_ICON_FAIL, "Error tagging \"%s\"", file_name);

			if (res == AF_ERR_FILE)
				pm_printf(PM_TAG_PROGRESS, PM_ICON_NONE, "%s (errno %d)", strerror(save_errno), save_errno);
			else 
				pm_printf(PM_TAG_PROGRESS, PM_ICON_NONE, "Unknown error (%d)", res);

			goto cont;
		}
		pm_printf(PM_TAG_PROGRESS, PM_ICON_OK, "Tagged \"%s\"", file_name);
		count_tagged++;

	cont:
		if (af) {
			audio_file_delete(af);
			af = NULL;
		}
	}

	pm_printf(PM_TAG_PROGRESS, PM_ICON_INFO, "Done (tagged %d of %d files)", count_tagged, count_total);
	pm_scroll_to_bottom(PM_TAG_PROGRESS);
	pm_end(PM_TAG_PROGRESS);

	if (pi != NULL)
		free_parse_info(pi);
}


static void update_interface_fname()
{
	static gboolean last_title = FALSE;
	static gboolean last_artist = FALSE;
	static gboolean last_album = FALSE;
	static gboolean last_year = FALSE;
	static gboolean last_comment = FALSE;
	static gboolean last_track = FALSE;
	static gboolean last_genre = FALSE;

	gchar *format = gtk_entry_get_text(ent_tag_format);
	gboolean value;
	gboolean save;

	/* inhibit proccessing signals in 'cb_filed_changed' */
	save = ignore_field_changed;
	ignore_field_changed = TRUE;

	if (strstr(format, "<title>")) from_fn_title = TRUE;
	else from_fn_title = FALSE;
	if (strstr(format, "<artist>")) from_fn_artist = TRUE;
	else from_fn_artist = FALSE;
	if (strstr(format, "<album>")) from_fn_album = TRUE;
	else from_fn_album = FALSE;
	if (strstr(format, "<year>")) from_fn_year = TRUE;
	else from_fn_year = FALSE;
	if (strstr(format, "<comment>")) from_fn_comment = TRUE;
	else from_fn_comment = FALSE;
	if (strstr(format, "<track>")) from_fn_track = TRUE;
	else from_fn_track = FALSE;
	if (strstr(format, "<genre>")) from_fn_genre = TRUE;
	else from_fn_genre = FALSE;

	value = *use_filename && from_fn_title;
	if (last_title != value) {
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(cb_title), value);
		gtk_entry_set_text(ent_title, (value ? "<from filename>" : ""));
		gtk_widget_set_sensitive(GTK_WIDGET(cb_title), !value);
		gtk_widget_set_sensitive(GTK_WIDGET(ent_title), !value);
		last_title = value;
	}
	value = *use_filename && from_fn_artist;
	if (last_artist != value) {
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(cb_artist), value);
		gtk_entry_set_text(ent_artist, (value ? "<from filename>" : ""));
		gtk_widget_set_sensitive(GTK_WIDGET(cb_artist), !value);
		gtk_widget_set_sensitive(GTK_WIDGET(ent_artist), !value);
		last_artist = value;
	}
	value = *use_filename && from_fn_album;
	if (last_album != value) {
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(cb_album), value);
		gtk_entry_set_text(ent_album, (value ? "<from filename>" : ""));
		gtk_widget_set_sensitive(GTK_WIDGET(cb_album), !value);
		gtk_widget_set_sensitive(GTK_WIDGET(ent_album), !value);
		last_album = value;
	}
	value = *use_filename && from_fn_year;
	if (last_year != value) {
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(cb_year), value);
		gtk_entry_set_text(ent_year, (value ? "<fn>" : ""));
		gtk_widget_set_sensitive(GTK_WIDGET(cb_year), !value);
		gtk_widget_set_sensitive(GTK_WIDGET(ent_year), !value);
		last_year = value;
	}
	value = *use_filename && from_fn_comment;
	if (last_comment != value) {
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(cb_comment), value);
		gtk_entry_set_text(ent_comment, (value ? "<from filename>" : ""));
		gtk_widget_set_sensitive(GTK_WIDGET(cb_comment), !value);
		gtk_widget_set_sensitive(GTK_WIDGET(ent_comment), !value);
		last_comment = value;
	}
	value = *use_filename && from_fn_track;
	if (last_track != value) {
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(cb_track), value);
		gtk_entry_set_text(ent_track, (value ? "<fn>" : ""));
		gtk_widget_set_sensitive(GTK_WIDGET(cb_track), !value);
		gtk_widget_set_sensitive(GTK_WIDGET(ent_track), !value);
		if (value)
			gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(cb_track_auto), FALSE);
		gtk_widget_set_sensitive(GTK_WIDGET(cb_track_auto), !value);
		last_track = value;
	}
	value = *use_filename && from_fn_genre;
	if (last_genre != value) {
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(cb_genre), value);
		gtk_entry_set_text(ent_genre, (value ? "<from filename>" : ""));
		gtk_widget_set_sensitive(GTK_WIDGET(cb_genre), !value);
		gtk_widget_set_sensitive(GTK_WIDGET(combo_genre), !value);
		last_genre = value;
	}

	ignore_field_changed = save;
}

static void update_interface_clear()
{
	gboolean use_fn = *use_filename;
	gboolean value = !gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(cb_remove_tags));

	gtk_widget_set_sensitive(GTK_WIDGET(combo_tag_format), value);
	gtk_widget_set_sensitive(GTK_WIDGET(b_tag_edit_format), value);
	gtk_widget_set_sensitive(GTK_WIDGET(cb_use_filename), value);

	gtk_widget_set_sensitive(GTK_WIDGET(cb_title), value);
	gtk_widget_set_sensitive(GTK_WIDGET(cb_artist), value);
	gtk_widget_set_sensitive(GTK_WIDGET(cb_album), value);
	gtk_widget_set_sensitive(GTK_WIDGET(cb_year), value);
	gtk_widget_set_sensitive(GTK_WIDGET(cb_comment), value);
	gtk_widget_set_sensitive(GTK_WIDGET(cb_track), value);
	gtk_widget_set_sensitive(GTK_WIDGET(cb_genre), value);

	gtk_widget_set_sensitive(GTK_WIDGET(ent_title), value && !(use_fn && from_fn_title));
	gtk_widget_set_sensitive(GTK_WIDGET(ent_artist), value && !(use_fn && from_fn_artist));
	gtk_widget_set_sensitive(GTK_WIDGET(ent_album), value && !(use_fn && from_fn_album));
	gtk_widget_set_sensitive(GTK_WIDGET(ent_year), value && !(use_fn && from_fn_year));
	gtk_widget_set_sensitive(GTK_WIDGET(ent_comment), value && !(use_fn && from_fn_comment));
	gtk_widget_set_sensitive(GTK_WIDGET(ent_track), value && !(use_fn && from_fn_track));
	gtk_widget_set_sensitive(GTK_WIDGET(cb_track_auto), value && !(use_fn && from_fn_track));
	gtk_widget_set_sensitive(GTK_WIDGET(combo_genre), value && !(use_fn && from_fn_genre));
}


/* UI callbacks */
void cb_update_filename(GtkWidget *widget, gpointer data)
{
	*use_filename = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(cb_use_filename));
	update_interface_fname();
}

void cb_toggle_remove_tags(GtkWidget *widget, gpointer data)
{
	update_interface_clear();
	if (!gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(cb_remove_tags)))
		update_interface_fname();
}

void cb_field_changed(GtkWidget *widget, gpointer data)
{
	if (ignore_field_changed)
		return;

	if ((void*)widget == (void*)ent_title)
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(cb_title), TRUE);
	else if ((void*)widget == (void*)ent_artist)
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(cb_artist), TRUE);
	else if ((void*)widget == (void*)ent_album)
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(cb_album), TRUE);
	else if ((void*)widget == (void*)ent_year)
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(cb_year), TRUE);
	else if ((void*)widget == (void*)ent_genre)
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(cb_genre), TRUE);
	else if ((void*)widget == (void*)ent_comment)
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(cb_comment), TRUE);
	else if ((void*)widget == (void*)ent_track || (void*)widget == (void*)cb_track_auto)
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(cb_track), TRUE);
}

void cb_check_changed(GtkToggleButton *widget, gpointer data)
{
	gboolean save;
	
	if (gtk_toggle_button_get_active(widget))
		return;

	/* inhibit proccessing signals in 'cb_field_changed' */
	save = ignore_field_changed;
	ignore_field_changed = TRUE;

	if ((void*)widget == (void*)cb_title)
		gtk_entry_set_text(ent_title, "");
	else if ((void*)widget == (void*)cb_artist)
		gtk_entry_set_text(ent_artist, "");
	else if ((void*)widget == (void*)cb_album)
		gtk_entry_set_text(ent_album, "");
	else if ((void*)widget == (void*)cb_year)
		gtk_entry_set_text(ent_year, "");
	else if ((void*)widget == (void*)cb_genre)
		gtk_entry_set_text(ent_genre, "");
	else if ((void*)widget == (void*)cb_comment)
		gtk_entry_set_text(ent_comment, "");
	else if ((void*)widget == (void*)cb_track) {
		gtk_entry_set_text(ent_track, "");
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(cb_track_auto), FALSE);
	}

	ignore_field_changed = save;
}

void cb_show_tag_chconv(GtkButton *button, gpointer user_data)
{
	chconv_display(CHCONV_TAG);
}

void cb_tag_help(GtkButton *button, gpointer user_data)
{
	help_display(HELP_TAG_FORMAT);
}

void cb_tag_go(GtkButton *button, gpointer user_data)
{
	GEList *list = ds_get_file_list();
	if (g_elist_length(list) == 0) {
		pm_start(PM_TAG_PROGRESS);
		pm_printf(PM_TAG_PROGRESS, PM_ICON_FAIL, "No files selected");
		pm_end(PM_TAG_PROGRESS);
		return;
	}

	cursor_set_wait();

	if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(cb_remove_tags))) {
		int button;
		button = modal_dialog(w_main, "Remove all tags", 
				      "This will remove tags from all selected files.",
				      2, 2, "Remove", "Cancel");
		if (button == 1)
			clear_tags();
	} else {
		mru_add(format_mru, gtk_entry_get_text(ent_tag_format));
		gtk_combo_set_popdown_strings(combo_tag_format, GLIST(format_mru->list));

		tag_files();
	}

	cursor_set_normal();
}


/*** public functions *******************************************************/

void tt_init(GladeXML *xml)
{
	GEList *genre_list;
	GEList *format_list;

	/*
	 * get the widgets from glade
	 */

	combo_tag_format = GTK_COMBO(glade_xml_get_widget(xml, "combo_tag_format"));
	ent_tag_format = GTK_ENTRY(glade_xml_get_widget(xml, "ent_tag_format"));
	b_tag_edit_format = GTK_BUTTON(glade_xml_get_widget(xml, "b_tag_edit_format"));
	cb_use_filename = GTK_CHECK_BUTTON(glade_xml_get_widget(xml, "cb_use_filename"));
	cb_remove_tags = GTK_CHECK_BUTTON(glade_xml_get_widget(xml, "cb_remove_tags"));
	b_tag_go = GTK_BUTTON(glade_xml_get_widget(xml, "b_tag_go"));

	cb_title = GTK_CHECK_BUTTON(glade_xml_get_widget(xml, "cb_title"));
	cb_artist = GTK_CHECK_BUTTON(glade_xml_get_widget(xml, "cb_artist"));
	cb_album = GTK_CHECK_BUTTON(glade_xml_get_widget(xml, "cb_album"));
	cb_year = GTK_CHECK_BUTTON(glade_xml_get_widget(xml, "cb_year"));
	cb_comment = GTK_CHECK_BUTTON(glade_xml_get_widget(xml, "cb_comment"));
	cb_track = GTK_CHECK_BUTTON(glade_xml_get_widget(xml, "cb_track"));
	cb_genre = GTK_CHECK_BUTTON(glade_xml_get_widget(xml, "cb_genre"));

	ent_title = GTK_ENTRY(glade_xml_get_widget(xml, "ent_title2"));
	ent_artist = GTK_ENTRY(glade_xml_get_widget(xml, "ent_artist2"));
	ent_album = GTK_ENTRY(glade_xml_get_widget(xml, "ent_album2"));
	ent_year = GTK_ENTRY(glade_xml_get_widget(xml, "ent_year2"));
	ent_comment = GTK_ENTRY(glade_xml_get_widget(xml, "ent_comment2"));
	ent_track = GTK_ENTRY(glade_xml_get_widget(xml, "ent_track2"));
	cb_track_auto = GTK_CHECK_BUTTON(glade_xml_get_widget(xml, "cb_track_auto"));
	combo_genre = GTK_COMBO(glade_xml_get_widget(xml, "combo_genre2"));
	ent_genre = GTK_ENTRY(glade_xml_get_widget(xml, "ent_genre2"));

	clist_tag_progress = GTK_WIDGET(glade_xml_get_widget(xml, "clist_tag_progress"));
	w_main = GTK_WINDOW(glade_xml_get_widget(xml, "w_main"));

	/* initialize some widgets' state */
	genre_list = genre_create_list(TRUE);
	g_elist_prepend(genre_list, "");
	gtk_combo_set_popdown_strings(combo_genre, GLIST(genre_list));
	g_elist_free(genre_list);


	/*
	 * get the preference values, or set them to defaults
	 */

	/* use_filename */
	use_filename = pref_get_ref("tt:use_filename");
	if (!use_filename) {
		gboolean temp = TRUE;
		use_filename = pref_set("tt:use_filename", PREF_BOOLEAN, &temp);
	}

	/* format_mru */
	format_list = pref_get_ref("tt:format_mru");
	if (!format_list) {
		GEList *temp_list = g_elist_new();
		g_elist_append(temp_list, "<track>. <title>");
		g_elist_append(temp_list, "<artist> - <title>");
		g_elist_append(temp_list, "<artist> - <album>/<track>. <title>");
		g_elist_append(temp_list, "<artist>/<album>/<track>. <title>");
		g_elist_append(temp_list, "<artist>/<album> (<year>)/<track>. <title>");
		format_list = pref_set("tt:format_mru", PREF_STRING | PREF_LIST, temp_list);
		g_elist_free(temp_list);
	}
	format_mru = mru_new_from_list(10, format_list);


	/*
	 * synchronize the interface state
	 */

	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(cb_use_filename), *use_filename);
	gtk_combo_set_popdown_strings(combo_tag_format, GLIST(format_mru->list));

	update_interface_fname();
}
